@@ -25,15 +25,19 @@ from nbconvert.preprocessors import ExecutePreprocessor
 from rich.console import Console
 from rich.progress import Progress
 
-from lineapy.data.types import SessionType
+from lineapy.api.models.linea_artifact import (
+    LineaArtifact,
+    get_lineaartifactdef,
+)
+from lineapy.data.types import ARTIFACT_STORAGE_BACKEND, SessionType
 from lineapy.db.db import RelationalLineaDB
 from lineapy.exceptions.excepthook import set_custom_excepthook
-from lineapy.graph_reader.apis import LineaArtifact
+from lineapy.graph_reader.artifact_collection import ArtifactCollection
 from lineapy.instrumentation.tracer import Tracer
-from lineapy.plugins.airflow import AirflowPlugin
+from lineapy.plugins.airflow_pipeline_writer import AirflowPipelineWriter
 from lineapy.plugins.utils import slugify
-from lineapy.transformer.node_transformer import transform
-from lineapy.utils.analytics import send_lib_info_from_db
+from lineapy.transformer.transform_code import transform
+from lineapy.utils.analytics.utils import send_lib_info_from_db
 from lineapy.utils.benchmarks import distribution_change
 from lineapy.utils.config import (
     CONFIG_FILE_NAME,
@@ -58,6 +62,13 @@ logger = logging.getLogger(__name__)
     help="Print out logging for graph creation and execution.",
     is_flag=True,
 )
+@click.version_option(
+    None,
+    "--version",
+    "-v",
+    message="%(package)s %(version)s",
+    help="Print the Lineapy version number and exit.",
+)
 @click.option(
     "--home-dir",
     type=click.Path(dir_okay=True, path_type=pathlib.Path),
@@ -82,22 +93,38 @@ logger = logging.getLogger(__name__)
     "--do-not-track",
     type=click.BOOL,
     help="Opt out for user analytics.",
-    is_flag=True,
 )
 @click.option(
     "--logging-level",
     type=click.Choice(
-        list(logging._nameToLevel.keys())
-        + sorted([str(x) for x in set(logging._levelToName.keys())]),
+        list(logging._nameToLevel.keys()),
         case_sensitive=False,
     ),
-    help="Logging level for LineaPy.",
+    help="Logging level for LineaPy, overrides the --verbose flag if set.",
 )
 @click.option(
     "--logging-file",
     type=click.Path(dir_okay=False, path_type=pathlib.Path),
     help="Logging file",
 )
+@click.option(
+    "--mlflow-registry-uri",
+    type=click.STRING,
+    help="MLFlow registry uri for ML models storage backend.",
+)
+@click.option(
+    "--mlflow-tracking-uri",
+    type=click.STRING,
+    help="MLFlow tracking uri for ML models storage backend.",
+)
+@click.option(
+    "--default-ml-models-storage-backend",
+    type=click.Choice(
+        [member.name for member in ARTIFACT_STORAGE_BACKEND],
+        case_sensitive=False,
+    ),
+    help="Default storage backend for ML models",
+)
 def linea_cli(
     verbose: bool,
     home_dir: Optional[pathlib.Path],
@@ -107,6 +134,9 @@ def linea_cli(
     do_not_track: Optional[bool],
     logging_level: Optional[str],
     logging_file: Optional[pathlib.Path],
+    mlflow_registry_uri: Optional[str],
+    mlflow_tracking_uri: Optional[str],
+    default_ml_models_storage_backend: Optional[ARTIFACT_STORAGE_BACKEND],
 ):
     """
     Pass all configuration to lineapy_config
@@ -115,7 +145,9 @@ def linea_cli(
 
     # Set the logging env variable so its passed to subprocesses, like creating a jupyter kernel
     if verbose:
-        options.set("LINEAPY_LOG_LEVEL", "DEBUG")
+        options.set("logging_level", "DEBUG")
+    if logging_level:
+        options.set("logging_level", logging_level)
 
     configure_logging()
 
@@ -123,8 +155,6 @@ def linea_cli(
         if arg in options.__dict__.keys() and locals().get(arg) is not None:
             options.set(arg, locals().get(arg))
 
-    options._set_defaults()
-
 
 @linea_cli.command()
 @click.option(
@@ -152,7 +182,7 @@ def init(output_file: Optional[pathlib.Path]):
         config = {
             k: str(v) for k, v in options.__dict__.items() if v is not None
         }
-        json.dump(config, f)
+        json.dump(config, f, indent=4, sort_keys=True)
 
 
 @linea_cli.command()
@@ -201,16 +231,17 @@ def notebook(
     # TODO: duplicated with `get` but no context set, should rewrite eventually
     # to not duplicate
     db = RelationalLineaDB.from_config(options)
-    artifact = db.get_artifact_by_name(artifact_name)
+    artifactorm = db.get_artifactorm_by_name(artifact_name)
     # FIXME: mypy issue with SQLAlchemy, see https://github.com/python/typeshed/issues/974
     api_artifact = LineaArtifact(
         db=db,
-        _execution_id=artifact.execution_id,
-        _node_id=artifact.node_id,
-        _session_id=artifact.node.session_id,
-        _version=artifact.version,  # type: ignore
+        _artifact_id=artifactorm.id,
+        _execution_id=artifactorm.execution_id,
+        _node_id=artifactorm.node_id,
+        _session_id=artifactorm.node.session_id,
+        _version=artifactorm.version,  # type: ignore
         name=artifact_name,
-        date_created=artifact.date_created,  # type: ignore
+        date_created=artifactorm.date_created,  # type: ignore
     )
     logger.info(api_artifact.get_code())
 
@@ -252,15 +283,16 @@ def file(
 
     # Print the slice:
     # FIXME: weird indirection
-    artifact = db.get_artifact_by_name(artifact_name)
+    artifactorm = db.get_artifactorm_by_name(artifact_name)
     api_artifact = LineaArtifact(
         db=db,
-        _execution_id=artifact.execution_id,
-        _node_id=artifact.node_id,
-        _session_id=artifact.node.session_id,
-        _version=artifact.version,  # type:ignore
+        _artifact_id=artifactorm.id,
+        _execution_id=artifactorm.execution_id,
+        _node_id=artifactorm.node_id,
+        _session_id=artifactorm.node.session_id,
+        _version=artifactorm.version,  # type:ignore
         name=artifact_name,
-        date_created=artifact.date_created,  # type:ignore
+        date_created=artifactorm.date_created,  # type:ignore
     )
     logger.info(api_artifact.get_code())
 
@@ -301,7 +333,12 @@ def generate_save_code(
     "--export-slice-to-airflow-dag",
     "--airflow",
     default=None,
-    help="Requires --slice. Export the sliced code from all slices to an Airflow DAG {export-slice-to-airflow-dag}.py",
+    help="Requires --slice. Export the sliced code from all slices to an Airflow DAG {export-dir}/{export-slice-to-airflow-dag}.py",
+)
+@click.option(
+    "--export-dir",
+    default=".",
+    help="Requires --slice. Export the sliced code from all slices to an Airflow DAG {export-dir}/{export-slice-to-airflow-dag}.py",
 )
 @click.option(
     "--airflow-task-dependencies",
@@ -336,13 +373,41 @@ def python(
     slice: List[str],  # cast tuple into list
     export_slice: List[str],  # cast tuple into list
     export_slice_to_airflow_dag: str,
+    export_dir: str,
     airflow_task_dependencies: str,
     print_source: bool,
     print_graph: bool,
     visualize: bool,
     arg: Iterable[str],
+):
+    python_cli(
+        file_name,
+        slice,
+        export_slice,
+        export_slice_to_airflow_dag,
+        export_dir,
+        airflow_task_dependencies,
+        print_source,
+        print_graph,
+        visualize,
+        arg,
+    )
+
+
+def python_cli(
+    file_name: pathlib.Path,
+    slice: List[str] = None,  # cast tuple into list
+    export_slice: List[str] = None,  # cast tuple into list
+    export_slice_to_airflow_dag: str = None,
+    export_dir: str = ".",
+    airflow_task_dependencies: str = None,
+    print_source: bool = False,
+    print_graph: bool = False,
+    visualize: bool = False,
+    arg: Iterable[str] = [],
 ):
     set_custom_excepthook()
+    check_python_version()
     tree = rich.tree.Tree(f"📄 {file_name}")
 
     db = RelationalLineaDB.from_config(options)
@@ -388,7 +453,9 @@ def python(
             exit(1)
         for _slice, _export_slice in zip(slice, export_slice):
             full_code = tracer.slice(_slice)
-            pathlib.Path(f"{_export_slice}.py").write_text(full_code)
+            (pathlib.Path(f"{export_dir}") / f"{_export_slice}.py").write_text(
+                full_code
+            )
 
     if export_slice_to_airflow_dag:
         if not slice:
@@ -397,13 +464,29 @@ def python(
             )
             exit(1)
 
-        ap = AirflowPlugin(db, tracer.tracer_context.get_session_id())
-        ap.sliced_airflow_dag(
-            slice,
-            export_slice_to_airflow_dag,
-            ast.literal_eval(airflow_task_dependencies),
+        # TODO: Use `Pipeline` object as an entry point (LIN-319 needs
+        # to be tackled first to define/refine expected behavior for CLI).
+        art_slice_defs = [
+            get_lineaartifactdef(art_entry=art_entry) for art_entry in slice
+        ]
+        task_dependencies = ast.literal_eval(airflow_task_dependencies or "{}")
+
+        artifact_collection = ArtifactCollection(
+            db,
+            art_slice_defs,
+            dependencies=task_dependencies,
         )
 
+        # Construct pipeline writer
+        pipeline_writer = AirflowPipelineWriter(
+            artifact_collection=artifact_collection,
+            pipeline_name=export_slice_to_airflow_dag,
+            output_dir=export_dir,
+        )
+
+        # Write out pipeline files
+        pipeline_writer.write_pipeline_files()
+
     db.close()
     if print_graph:
         graph_code = prettify(
@@ -455,6 +538,7 @@ def setup_ipython_dir() -> None:
     a line to add lineapy into ``extra_extensions``. If they do not exist, we create
     new config files in the temp folder and add a line to specify ``extra_extensions``.
     """
+    check_python_version()
     ipython_dir_name = tempfile.mkdtemp()
     # Make a default profile with the extension added to the ipython and kernel
     # configs
@@ -545,6 +629,11 @@ def add(path: pathlib.Path, name: str):
     This command copies the yaml file whose path is provided by the user into the user's .lineapy directory to allow Lineapy to manage it.
     """
 
+    annotations_add(path, name)
+
+
+def annotations_add(path: pathlib.Path, orig_name: Optional[str] = None):
+
     # validate that source is in correct format before importing
     try:
         invalid_specs = validate_spec(path)
@@ -557,7 +646,7 @@ def add(path: pathlib.Path, name: str):
         exit(1)
 
     # calculate name of new annotation source
-    name = name or path.stem
+    name = orig_name or path.stem
     name = remove_annotations_file_extension(name)
     name = slugify(name)
 
@@ -585,6 +674,10 @@ def list():
     """
     Lists full paths to all imported annotation sources.
     """
+    annotations_list()
+
+
+def annotations_list():
     wildcard_path = os.path.join(
         pathlib.Path(
             options.safe_get("customized_annotation_folder")
@@ -611,7 +704,12 @@ def delete(name: str):
     """
     Deletes imported annotation source.
     """
-    name = remove_annotations_file_extension(name)
+
+    annotations_delete(name)
+
+
+def annotations_delete(custom_name: str):
+    name = remove_annotations_file_extension(custom_name)
     name += CUSTOM_ANNOTATIONS_EXTENSION_NAME
 
     delete_path = pathlib.Path(
@@ -619,6 +717,9 @@ def delete(name: str):
     ).joinpath(name)
     try:
         os.remove(delete_path)
+        logger.info(
+            f"Deleted annotation source {custom_name} at {delete_path}"
+        )
     except IsADirectoryError as e:
         logger.error(
             f"{delete_path} must be the path to a file, not a directory\n{e}"
@@ -716,5 +817,13 @@ def benchmark(path: pathlib.Path, n: int, skip_baseline: bool):
         rich.print(f"Lineapy is {str(change)}")
 
 
+def check_python_version():
+    if sys.version_info >= (3, 11):
+        logger.warn(
+            "WARNING: Support for Python 3.11 and above is in the experimental phase. \
+Please use caution while using LineaPy on your current Python installation!"
+        )
+
+
 if __name__ == "__main__":
     linea_cli()
