@@ -1,10 +1,23 @@
 from dataclasses import dataclass
+from enum import Enum
 from typing import Optional, Union
 
 # I think amplitude doesn't really support nested objects,
 # so flattening the object here.
 
 
+class ErrorType(Enum):
+    SAVE = "save_API_exception"
+    DELETE = "delete_API_exception"
+    RETRIEVE = "retrieve_API_exception"
+    PIPELINE = "pipeline_API_exception"
+    DATABASE = "database_error"
+    USER = "user_exception"
+
+    def __str__(self):
+        return str(self.value)
+
+
 @dataclass
 class SaveEvent:
     side_effect: str  # "file_system", "db", or "" (for variables)
@@ -36,8 +49,7 @@ class LibImportEvent:
 
 @dataclass
 class ExceptionEvent:
-    # TODO: set error_type to be string enums
-    error_type: str
+    error_type: ErrorType
     error_msg: Optional[str] = None
 
 
@@ -57,7 +69,17 @@ class GetVersionEvent:
     dummy_entry: str  # just a dummy entry for now
 
 
-AllEvents = Union[
+@dataclass
+class CyclicGraphEvent:
+    dummy_entry: str  # dummy entry, may populate with extra information later
+
+
+@dataclass
+class TagEvent:
+    tag: str  # dummy entry, may populate with extra information later
+
+
+TrackingEvent = Union[
     CatalogEvent,
     LibImportEvent,
     ExceptionEvent,
@@ -67,4 +89,6 @@ AllEvents = Union[
     GetCodeEvent,
     GetValueEvent,
     GetVersionEvent,
+    CyclicGraphEvent,
+    TagEvent,
 ]
